@@ -7,9 +7,10 @@ import (
 	"crypto/tls"
 	"crypto/x509"
 	"encoding/asn1"
+	"encoding/base32"
 	"encoding/base64"
 	"fmt"
-	"io/ioutil"
+	"io"
 	"log"
 	"math/rand"
 	"net"
@@ -26,6 +27,7 @@ import (
 	"github.com/letsencrypt/challtestsrv"
 	"github.com/letsencrypt/pebble/v2/acme"
 	"github.com/letsencrypt/pebble/v2/core"
+	"github.com/letsencrypt/pebble/v2/db"
 )
 
 const (
@@ -92,6 +94,8 @@ type vaTask struct {
 	Identifier acme.Identifier
 	Challenge  *core.Challenge
 	Account    *core.Account
+	AccountURL string
+	Wildcard   bool
 }
 
 type VAImpl struct {
@@ -105,12 +109,19 @@ type VAImpl struct {
 	strict             bool
 	customResolverAddr string
 	dnsClient          *dns.Client
+
+	// The VA having a DB client is indeed strange. This is only used to
+	// facilitate va.setOrderError changing the ARI related order replacement
+	// field on failed orders.
+	db *db.MemoryStore
 }
 
 func New(
 	log *log.Logger,
 	httpPort, tlsPort int,
-	strict bool, customResolverAddr string) *VAImpl {
+	strict bool, customResolverAddr string,
+	db *db.MemoryStore,
+) *VAImpl {
 	va := &VAImpl{
 		log:                log,
 		httpPort:           httpPort,
@@ -120,6 +131,7 @@ func New(
 		sleepTime:          defaultSleepTime,
 		strict:             strict,
 		customResolverAddr: customResolverAddr,
+		db:                 db,
 	}
 
 	if customResolverAddr != "" {
@@ -156,11 +168,13 @@ func New(
 	return va
 }
 
-func (va VAImpl) ValidateChallenge(ident acme.Identifier, chal *core.Challenge, acct *core.Account) {
+func (va VAImpl) ValidateChallenge(ident acme.Identifier, chal *core.Challenge, acct *core.Account, acctURL string, wildcard bool) {
 	task := &vaTask{
 		Identifier: ident,
 		Challenge:  chal,
 		Account:    acct,
+		AccountURL: acctURL,
+		Wildcard:   wildcard,
 	}
 	// Submit the task for validation
 	va.tasks <- task
@@ -203,10 +217,17 @@ func (va VAImpl) setAuthzValid(authz *core.Authorization, chal *core.Challenge)
 
 // setOrderError updates an order with an error from an authorization
 // validation.
-func (va VAImpl) setOrderError(order *core.Order, err *acme.ProblemDetails) {
+func (va VAImpl) setOrderError(order *core.Order, prob *acme.ProblemDetails) {
 	order.Lock()
 	defer order.Unlock()
-	order.Error = err
+	order.Error = prob
+
+	// Mark the parent order as "not replaced yet" so a new replacement order
+	// can be attempted.
+	err := va.db.UpdateReplacedOrder(order.Replaces, false)
+	if err != nil {
+		va.log.Printf("Error updating replacement order: %s", err)
+	}
 }
 
 // setAuthzInvalid updates an authorization and an associated challenge to be
@@ -215,7 +236,8 @@ func (va VAImpl) setOrderError(order *core.Order, err *acme.ProblemDetails) {
 func (va VAImpl) setAuthzInvalid(
 	authz *core.Authorization,
 	chal *core.Challenge,
-	err *acme.ProblemDetails) {
+	err *acme.ProblemDetails,
+) {
 	authz.Lock()
 	defer authz.Unlock()
 	// Update the authz status
@@ -268,9 +290,9 @@ func (va VAImpl) process(task *vaTask) {
 func (va VAImpl) performValidation(task *vaTask, results chan<- *core.ValidationRecord) {
 	if va.sleep {
 		// Sleep for a random amount of time between 0 and va.sleepTime seconds
-		len := time.Duration(rand.Intn(va.sleepTime))
-		va.log.Printf("Sleeping for %s seconds before validating", time.Second*len)
-		time.Sleep(time.Second * len)
+		length := time.Duration(rand.Intn(va.sleepTime)) * time.Second
+		va.log.Printf("Sleeping for %s seconds before validating", length)
+		time.Sleep(length)
 	}
 
 	// If `alwaysValid` is true then return a validation record immediately
@@ -297,6 +319,8 @@ func (va VAImpl) performValidation(task *vaTask, results chan<- *core.Validation
 		results <- va.validateTLSALPN01(task)
 	case acme.ChallengeDNS01:
 		results <- va.validateDNS01(task)
+	case acme.ChallengeDNSAccount01:
+		results <- va.validateDNSAccount01(task)
 	default:
 		va.log.Printf("Error: performValidation(): Invalid challenge type: %q", task.Challenge.Type)
 	}
@@ -318,7 +342,7 @@ func (va VAImpl) validateDNS01(task *vaTask) *core.ValidationRecord {
 	}
 
 	if len(txts) == 0 {
-		msg := fmt.Sprintf("No TXT records found for DNS challenge")
+		msg := "No TXT records found for DNS challenge"
 		result.Error = acme.UnauthorizedProblem(msg)
 		return result
 	}
@@ -335,7 +359,50 @@ func (va VAImpl) validateDNS01(task *vaTask) *core.ValidationRecord {
 		}
 	}
 
-	msg := fmt.Sprintf("Correct value not found for DNS challenge")
+	msg := "Correct value not found for DNS challenge"
+	result.Error = acme.UnauthorizedProblem(msg)
+	return result
+}
+
+func (va VAImpl) validateDNSAccount01(task *vaTask) *core.ValidationRecord {
+	acctHash := sha256.Sum256([]byte(task.AccountURL))
+	acctLabel := strings.ToLower(base32.StdEncoding.EncodeToString(acctHash[0:10]))
+	scope := "host"
+	if task.Wildcard {
+		scope = "wildcard"
+	}
+	challengeSubdomain := fmt.Sprintf("_%s._acme-%s-challenge.%s", acctLabel, scope, task.Identifier.Value)
+
+	result := &core.ValidationRecord{
+		URL:         challengeSubdomain,
+		ValidatedAt: time.Now(),
+	}
+
+	txts, err := va.getTXTEntry(challengeSubdomain)
+	if err != nil {
+		result.Error = acme.UnauthorizedProblem(fmt.Sprintf("Error retrieving TXT records for DNS-ACCOUNT-01 challenge (%q)", err))
+		return result
+	}
+
+	if len(txts) == 0 {
+		msg := "No TXT records found for DNS-ACCOUNT-01 challenge"
+		result.Error = acme.UnauthorizedProblem(msg)
+		return result
+	}
+
+	task.Challenge.RLock()
+	expectedKeyAuthorization := task.Challenge.ExpectedKeyAuthorization(task.Account.Key)
+	h := sha256.Sum256([]byte(expectedKeyAuthorization))
+	task.Challenge.RUnlock()
+	authorizedKeysDigest := base64.RawURLEncoding.EncodeToString(h[:])
+
+	for _, element := range txts {
+		if subtle.ConstantTimeCompare([]byte(element), []byte(authorizedKeysDigest)) == 1 {
+			return result
+		}
+	}
+
+	msg := "Correct value not found for DNS-ACCOUNT-01 challenge"
 	result.Error = acme.UnauthorizedProblem(msg)
 	return result
 }
@@ -356,7 +423,6 @@ func (va VAImpl) validateTLSALPN01(task *vaTask) *core.ValidationRecord {
 	}
 
 	addrs, err := va.resolveIP(task.Identifier.Value)
-
 	if err != nil {
 		result.Error = acme.MalformedProblem(
 			fmt.Sprintf("Error occurred while resolving URL %q: %q", task.Identifier.Value, err))
@@ -451,7 +517,6 @@ func (va VAImpl) validateTLSALPN01(task *vaTask) *core.ValidationRecord {
 
 func (va VAImpl) fetchConnectionState(hostPort string, config *tls.Config) (*tls.ConnectionState, *acme.ProblemDetails) {
 	conn, err := tls.DialWithDialer(&net.Dialer{Timeout: validationTimeout}, "tcp", hostPort, config)
-
 	if err != nil {
 		// TODO(@cpu): Return better err - see parseHTTPConnError from boulder
 		return nil, acme.UnauthorizedProblem(
@@ -505,7 +570,7 @@ func (va VAImpl) fetchHTTP(identifier string, token string) ([]byte, string, *ac
 	}
 
 	va.log.Printf("Attempting to validate w/ HTTP: %s\n", url)
-	httpRequest, err := http.NewRequest("GET", url.String(), nil)
+	httpRequest, err := http.NewRequest(http.MethodGet, url.String(), nil)
 	if err != nil {
 		return nil, url.String(), acme.MalformedProblem(
 			fmt.Sprintf("Invalid URL %q\n", url.String()))
@@ -535,7 +600,7 @@ func (va VAImpl) fetchHTTP(identifier string, token string) ([]byte, string, *ac
 			// Control specifically which IP will be used for this request
 			addrs, err := va.resolveIP(host)
 			if err != nil {
-				return nil, fmt.Errorf("error occurred while resolving URL %q: %q", url.String(), err)
+				return nil, fmt.Errorf("error occurred while resolving URL %q: %w", url.String(), err)
 			}
 			if len(addrs) == 0 {
 				return nil, fmt.Errorf("could not resolve URL %q", url.String())
@@ -558,7 +623,7 @@ func (va VAImpl) fetchHTTP(identifier string, token string) ([]byte, string, *ac
 	// NOTE: This is *not* using a `io.LimitedReader` and isn't suitable for
 	// production because a very large response will bog down the server. Don't
 	// use Pebble anywhere that isn't a testing rig!!!
-	body, err := ioutil.ReadAll(resp.Body)
+	body, err := io.ReadAll(resp.Body)
 	if err != nil {
 		return nil, url.String(), acme.InternalErrorProblem(err.Error())
 	}
@@ -567,7 +632,7 @@ func (va VAImpl) fetchHTTP(identifier string, token string) ([]byte, string, *ac
 		return nil, url.String(), acme.InternalErrorProblem(err.Error())
 	}
 
-	if resp.StatusCode != 200 {
+	if resp.StatusCode != http.StatusOK {
 		return nil, url.String(), acme.UnauthorizedProblem(
 			fmt.Sprintf("Non-200 status code from HTTP: %s returned %d",
 				url.String(), resp.StatusCode))
@@ -590,7 +655,6 @@ func (va VAImpl) getTXTEntry(name string) ([]string, error) {
 	message := new(dns.Msg)
 	message.SetQuestion(dns.Fqdn(name), dns.TypeTXT)
 	in, _, err := va.dnsClient.ExchangeContext(ctx, message, va.customResolverAddr)
-
 	if err != nil {
 		return nil, err
 	}
@@ -629,7 +693,6 @@ func (va VAImpl) resolveIP(name string) ([]string, error) {
 	messageAAAA := new(dns.Msg)
 	messageAAAA.SetQuestion(dns.Fqdn(name), dns.TypeAAAA)
 	inAAAA, _, err := va.dnsClient.ExchangeContext(ctx, messageAAAA, va.customResolverAddr)
-
 	if err != nil {
 		return nil, err
 	}
@@ -643,7 +706,6 @@ func (va VAImpl) resolveIP(name string) ([]string, error) {
 	messageA := new(dns.Msg)
 	messageA.SetQuestion(dns.Fqdn(name), dns.TypeA)
 	inA, _, err := va.dnsClient.ExchangeContext(ctx, messageA, va.customResolverAddr)
-
 	if err != nil {
 		return nil, err
 	}
